;;; test.scm --- MicroBenchmark library.
;;
;; Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
;;
;; This file is part of MicroBenchmark.
;;
;; MicroBenchmark is free software; you can redistribute it and/or
;; modify it under the terms of the GNU Lesser General Public License
;; as published by the Free Software Foundation; either version 3 of
;; the License, or (at your option) any later version.
;;
;; MicroBenchmark is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; Lesser General Public License for more details.
;;
;; You should have received a copy of the GNU Lesser General Public
;; License along with MicroBenchmark.  If not, see
;; <http://www.gnu.org/licenses/>.
;;
;;; Commentary:
;;
;; Test case definition.
;;
;;; Code:
(define-module (mbenchmark test)
  #:use-module (mbenchmark c test)
  #:use-module (mbenchmark suite)
  #:use-module (mbenchmark state)
  #:use-module (srfi srfi-9)
  #:export (register-test!
            add-test!
            test-case?

            set-constraints!
            add-dimension!
            dimensions
            skip-iterations!
            iterations-to-skip
            limit-iterations!
            min-iterations
            max-iterations
            limit-samples!
            max-sample-iterations
            min-sample-iterations))

(define-record-type <test-case>
  (make-test-case suite test)
  test-case?
  (suite test-case-suite)
  (test test-case-c-object))

(define (default-set-up state) (state-sizes state))
(define (default-tear-down . args) #t)

(define* (set-constraints-int test
                              #:key (dimensions #f)
                              (skip-iterations #f)
                              (min-iterations #f)
                              (max-iterations #f)
                              (min-sample-iterations #f)
                              (max-sample-iterations #f))
  (when dimensions
    (for-each (lambda (sizes)
                (add-dimension!/internal test sizes))
              dimensions))

  (when skip-iterations
    (skip-iterations!/internal test skip-iterations))

  (when (or min-iterations max-iterations)
    (let ((min (or min-iterations 0))
          (max (or max-iterations 0)))
      (limit-iterations!/internal test min max)))

  (when (or min-sample-iterations max-sample-iterations)
    (let ((min (or min-sample-iterations 0))
          (max (or max-sample-iterations 0)))
      (limit-samples!/internal test min max))))


(define (set-constraints! test . args)
  (apply set-constraints-int (test-case-c-object test) args))

(define (make-test-function f auto)
  (define (auto-function state args)
    (when (keep-running? state)
      (apply f args)
      (auto-function state args)))
  (define (directed-function state args)
    (apply f state args))
  (if auto auto-function directed-function))

(define* (register-test! name
                         #:key test
                         (auto #t)
                         (set-up default-set-up)
                         (tear-down default-tear-down)
                         (dimensions #f)
                         (skip-iterations #f)
                         (min-iterations #f)
                         (max-iterations #f)
                         (min-sample-iterations #f)
                         (max-sample-iterations #f))

  (define (apply-constraints tc)
    (set-constraints-int tc
                         #:dimensions dimensions
                         #:skip-iterations skip-iterations
                         #:min-iterations min-iterations
                         #:max-iterations max-iterations
                         #:min-sample-iterations min-sample-iterations
                         #:max-sample-iterations max-sample-iterations))

  (register-test!/internal name
                           (make-test-function test auto)
                           (if set-up set-up default-set-up)
                           (if tear-down tear-down default-tear-down))
  (register-constraint!/internal name apply-constraints))

(define* (add-test! suite name
                    #:key test
                    (auto #t)
                    (set-up default-set-up)
                    (tear-down default-tear-down)
                    (dimensions #f)
                    (skip-iterations #f)
                    (min-iterations #f)
                    (max-iterations #f)
                    (min-sample-iterations #f)
                    (max-sample-iterations #f))
  (let ((c-test (add-test!/internal suite name
                                    (make-test-function test auto)
                                    (or set-up default-set-up)
                                    (or tear-down default-tear-down))))
    (set-constraints-int c-test
                         #:dimensions dimensions
                         #:skip-iterations skip-iterations
                         #:min-iterations min-iterations
                         #:max-iterations max-iterations
                         #:min-sample-iterations min-sample-iterations
                         #:max-sample-iterations max-sample-iterations)
    (make-test-case suite c-test)))

(define (add-dimension! test sizes)
  (add-dimension!/internal (test-case-c-object test) sizes))

(define (dimensions test)
  (dimensions/internal (test-case-c-object test)))

(define (skip-iterations! test it)
  (skip-iterations!/internal (test-case-c-object test) it))

(define (iterations-to-skip test)
  (iterations-to-skip/internal (test-case-c-object test)))

(define* (limit-iterations! test #:key (min 0) (max 0))
  (limit-iterations!/internal (test-case-c-object test) min max))

(define (max-iterations test)
  (max-iterations/internal (test-case-c-object test)))

(define (min-iterations test)
  (min-iterations/internal (test-case-c-object test)))

(define* (limit-samples! test #:key (min-iterations 0) (max-iterations 0))
  (limit-iterations!/internal (test-case-c-object test) min-iterations
                              max-iterations))

(define (max-sample-iterations test)
  (max-sample-iterations/internal (test-case-c-object test)))

(define (min-sample-iterations test)
  (min-sample-iterations/internal (test-case-c-object test)))

