/* test.hpp --- MicroBenchmark C++ library. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifndef MICRO_BENCHMARK_CXX_TEST_HPP
#define MICRO_BENCHMARK_CXX_TEST_HPP

#include "mbenchmark/c++/c-handler.hpp"
#include "mbenchmark/c++/state.hpp"
#include "mbenchmark/c++/traits.hpp"

#include <cstddef>
#include <initializer_list>
#include <memory>
#include <vector>
#include <type_traits>
#include <utility>

namespace micro_benchmark
{
  MICRO_BENCHMARK_BEGIN_VERSION_NAMESPACE
  namespace detail
  {
    struct default_set_up
    {
      std::vector<std::size_t>
      operator() (state const& state) const
      {
        return state.sizes ();
      }
    };
    constexpr default_set_up set_up{};

    struct default_tear_down
    {
      template <typename Data>
      void
      operator() (state const&, Data&&) const
      {
      }
    };
    constexpr default_tear_down tear_down{};

    class test_base
    {
    protected:
      test_base () = default;
    public:
      virtual ~test_base () = default;
      test_base (test_base const&) = delete;
      test_base& operator= (test_base const&) = delete;
      test_base (test_base&&) = delete;
      test_base& operator= (test_base&&) = delete;

      virtual void run_set_up (state const& state) noexcept = 0;
      virtual void run_tear_down (state const& state) noexcept = 0;
      virtual void run_test (state const& state) const noexcept = 0;
    };

    /* TODO: Directed-test.  */
    template <typename Test, typename SetUp, typename TearDown>
    class test_data final : public detail::test_base
    {

      Test test;
      SetUp set_up;
      TearDown tear_down;
      using SetUpRet =
        decltype (detail::as_ref<SetUp> () (detail::as_cref<state> ()));
      using SetUpData = detail::value_type<SetUpRet>;
      SetUpData data;

    public:
      explicit test_data (Test t, SetUp s, TearDown td)
        : test (std::move (t)),
          set_up (std::move (s)),
          tear_down (std::move (td)),
          data ()
      { }

      ~test_data () final = default;

      void
      run_set_up (state const& state) noexcept final
      {
        data = set_up (state);
      }

      void
      run_tear_down (state const& state) noexcept final
      {
        tear_down (state, data);
      }

      void
      run_test (state const& state) const noexcept final
      {
        while (state.keep_running ())
          test (data);
      }
    };

    template <typename T>
    struct add_pointer_to_function
    {
      using type = T;
    };

    template <typename T, typename... Args>
    struct add_pointer_to_function<T (Args...)>
    {
      using type = T (*) (Args...);
    };

    test_case make_test_case (void*, c_suite_ptr const&);

    template <typename Test, typename SetUp, typename TearDown>
    std::unique_ptr<test_base>
    make_test_data (Test&& test, SetUp&& set_up, TearDown&& tear_down)
    {
      using VT = typename detail::value_type<Test>;
      using T = typename add_pointer_to_function<VT>::type;

      using VS = typename detail::value_type<SetUp>;
      using S = typename add_pointer_to_function<VS>::type;

      using VD = typename detail::value_type<TearDown>;
      using D = typename add_pointer_to_function<VD>::type;

      using type = test_data<T, S, D>;
      using ptr_type = std::unique_ptr<test_base>;
      return ptr_type (new type {std::forward<Test> (test),
                                 std::forward<SetUp> (set_up),
                                 std::forward<TearDown> (tear_down)});
    }
  }

  class test_case : detail::c_handler
  {
    friend class suite;
    friend test_case detail::make_test_case (void*, detail::c_suite_ptr const&);
    detail::c_suite_ptr suite;

    test_case (detail::c_handler &&hnd, detail::c_suite_ptr const& suite)
      : detail::c_handler (std::move (hnd)), suite (suite)
    { }

    void register_data_ (std::unique_ptr<detail::test_base>&& base);

    template <typename Test, typename SetUp, typename TearDown>
    void
    register_data_ (Test&& test, SetUp&& set_up, TearDown&& tear_down)
    {
      auto ptr = detail::make_test_data (std::forward<Test> (test),
                                         std::forward<SetUp> (set_up),
                                         std::forward<TearDown> (tear_down));
      register_data_ (std::move (ptr));
    }

  public:
    ~test_case () = default;
    test_case (test_case&&) = default;
    test_case& operator= (test_case&&) = default;

    template <typename Range>
    void add_dimension (Range dimension)
    {
      using std::begin;
      using std::end;
      add_dimension (begin (dimension), end (dimension));
    }

    template <typename T>
    void add_dimension (std::initializer_list<T> dimension)
    {
      using std::begin;
      using std::end;
      add_dimension (begin (dimension), end (dimension));
    }

    template <typename Iterator>
    void add_dimension (Iterator begin, Iterator end)
    {
      std::vector<std::size_t> dim (begin, end);
      add_dimension (dim);
    }

    void add_dimension (std::vector<std::size_t> const& dim);

    std::size_t number_of_dimensions () const;
    std::vector<std::size_t> get_dimension (std::size_t) const;

    void skip_iterations (std::size_t iterations);
    std::size_t iterations_to_skip () const;

    void limit_iterations (std::size_t min, std::size_t max);
    std::size_t min_iterations () const;
    std::size_t max_iterations () const;

    void limit_samples (std::size_t min, std::size_t max);
    std::size_t min_sample_iterations () const;
    std::size_t max_sample_iterations () const;
  };
  MICRO_BENCHMARK_END_VERSION_NAMESPACE
}
#endif
