/* console.c --- MicroBenchmark library. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "internal/output/console.h"

#include "mbenchmark/stats.h"
#include "mbenchmark/report/exec.h"
#include "internal/output/common.h"
#include "internal/utility/log.h"
#include "internal/utility/error.h"
#include "internal/utility/intl.h"
#include "internal/utility/xalloc.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#define MICRO_BENCHMARK_SUBMODULE "console"
#define MICRO_BENCHMARK_MODULE "output:" MICRO_BENCHMARK_SUBMODULE

#define SEPARATOR_EQ_36 "===================================="
#define SEPARATOR_EQ_72 SEPARATOR_EQ_36 SEPARATOR_EQ_36 "\n"

#define SEPARATOR_DS_36 "------------------------------------"
#define SEPARATOR_DS_72 SEPARATOR_DS_36 SEPARATOR_DS_36 "\n"


static const char *
stat_value_to_string (micro_benchmark_output_stat v)
{
  switch ((int) v)
    {
    case MICRO_BENCHMARK_STAT_ALL:
      return _("μ/σ²/σ");

    case MICRO_BENCHMARK_STAT_BASIC:
      return _("μ/σ");

    case MICRO_BENCHMARK_STAT_MEAN | MICRO_BENCHMARK_STAT_VARIANCE:
      return _("μ/σ²");

    case MICRO_BENCHMARK_STAT_VARIANCE | MICRO_BENCHMARK_STAT_STD_DEVIATION:
      return _("σ²/σ");

    case MICRO_BENCHMARK_STAT_MEAN:
      return _("μ");

    case MICRO_BENCHMARK_STAT_VARIANCE:
      return _("σ²");

    case MICRO_BENCHMARK_STAT_STD_DEVIATION:
      return _("σ");

    default:
      MB_HANDLE_ERROR (true, "Cannot convert stat value to string");
      return "";
    }
}

/*  TODO: Real table format.  */
static const int test_name_size = 20;
static const int field_min_size = 10;

static void
write_header (FILE *out, const micro_benchmark_output_values *v,
              const char *name, size_t num_tests)
{
  fprintf (out, _("Suite: %s (%zu tests)\n"), name, num_tests);
  fprintf (out, SEPARATOR_EQ_72);
  fprintf (out, "%*s", test_name_size, _("Test Name"));
  if (v->total_iterations)
    fprintf (out, "\t%-*s", field_min_size, _("Iter.Total"));

  if (v->iterations)
    fprintf (out, "\t%-*s", field_min_size, _("Iterations"));

  if (v->iteration_time)
    fprintf (out, "\t%-*s(%s)", field_min_size - 3, _("It.Time"),
             stat_value_to_string (v->iteration_time));

  if (v->total_samples)
    fprintf (out, "\t%-*s", field_min_size, _("Samp.Total"));

  if (v->samples)
    fprintf (out, "\t%-*s", field_min_size, _("Samples"));

  if (v->sample_time)
    fprintf (out, "\t%-*s(%s)", field_min_size - 3, _("Samp. Time"),
             stat_value_to_string (v->sample_time));

  if (v->sample_iterations)
    fprintf (out, "\t%-*s(%s)", field_min_size - 3, _("Samp.Iter."),
             stat_value_to_string (v->sample_iterations));

  if (v->total_time)
    fprintf (out, "\t%-*s", field_min_size, _("Total Time"));

  if (v->size_constraints)
    fprintf (out, "\t%-*s", field_min_size, _("Sizes"));

  fprintf (out, "\n");
  fprintf (out, SEPARATOR_EQ_72);
}

static void
print_time_value (FILE *out, micro_benchmark_stats_value v,
                  micro_benchmark_output_stat stat)
{
  static const int precision = 4;
  const char *units = micro_benchmark_time_units_to_string_ (v.unit);
  const char *unitssqrd = micro_benchmark_time_units_sqrd_to_string_ (v.unit);

  fprintf (out, "\t");
  switch ((int) stat)
    {
    default:
      assert (0);

    case MICRO_BENCHMARK_STAT_ALL:
      fprintf (out, _("%.*f%s/%.*f%s/%.*f%s"), precision / 4, v.mean, units,
               precision / 4, v.variance, unitssqrd,
               precision / 4, v.std_deviation, units);
      break;

    case MICRO_BENCHMARK_STAT_BASIC:
      fprintf (out, _("%.*f%s/%.*f%s"), precision / 4, v.mean, units,
               precision / 4, v.std_deviation, units);
      break;

    case MICRO_BENCHMARK_STAT_MEAN | MICRO_BENCHMARK_STAT_VARIANCE:
      fprintf (out, _("%.*f%s/%.*f%s"), precision / 4, v.mean, units,
               precision / 4, v.variance, unitssqrd);
      break;

    case MICRO_BENCHMARK_STAT_VARIANCE | MICRO_BENCHMARK_STAT_STD_DEVIATION:
      fprintf (out, _("%.*f%s/%.*f%s"), precision / 4, v.variance, unitssqrd,
               precision / 4, v.std_deviation, units);
      break;

    case MICRO_BENCHMARK_STAT_STD_DEVIATION:
      fprintf (out, _("%.*f%s"), precision, v.std_deviation, units);
      break;

    case MICRO_BENCHMARK_STAT_VARIANCE:
      fprintf (out, _("%.*f%s"), precision, v.variance, unitssqrd);
      break;

    case MICRO_BENCHMARK_STAT_MEAN:
      fprintf (out, _("%.*f%s"), precision, v.mean, units);
      break;
    }
}

static void
write_execution_report (micro_benchmark_exec_report report, FILE *out,
                        const micro_benchmark_output_values *values)
{
  const char *full_name = micro_benchmark_exec_report_get_name (report);
  fprintf (out, "%*s", test_name_size, full_name);

  if (values->total_iterations)
    {
      size_t it = micro_benchmark_exec_report_total_iterations (report);
      fprintf (out, "\t%-*zu", field_min_size, it);
    }

  if (values->iterations)
    {
      size_t it = micro_benchmark_exec_report_get_iterations (report);
      fprintf (out, "\t%-*zu", field_min_size, it);
    }

  if (values->iteration_time)
    {
      micro_benchmark_stats_value it_time =
        micro_benchmark_exec_report_iteration_time (report);
      print_time_value (out, it_time, values->iteration_time);
    }

  if (values->total_samples)
    {
      size_t total = micro_benchmark_exec_report_total_samples (report);
      fprintf (out, "\t%-*zu", field_min_size, total);
    }

  if (values->samples)
    {
      size_t samples = micro_benchmark_exec_report_used_samples (report);
      fprintf (out, "\t%-*zu", field_min_size, samples);
    }

  if (values->sample_time)
    {
      micro_benchmark_stats_value s_time =
        micro_benchmark_exec_report_sample_time (report);
      print_time_value (out, s_time, values->sample_time);
    }

  if (values->sample_iterations)
    {
      micro_benchmark_stats_value s_it =
        micro_benchmark_exec_report_sample_iterations (report);
      print_time_value (out, s_it, values->sample_iterations);
    }

  if (values->total_time)
    {
      micro_benchmark_clock_time elapsed =
        micro_benchmark_exec_report_total_time (report);
      /*  TODO: Extract time printing to its module  */
      fprintf (out, "\t%d.%06d%s",
               (int) elapsed.seconds,
               (int) elapsed.nanoseconds / 1000, _("s"));
    }

  if (values->size_constraints)
    {
      const size_t *sizes;
      size_t dimensions;
      micro_benchmark_exec_report_get_sizes (report, &sizes, &dimensions);
      if (dimensions > 0)
        {
          fprintf (out, "\t{%zu", sizes[0]);
          for (size_t d = 1; d < dimensions; ++d)
            fprintf (out, ",%zu", sizes[d]);
          fprintf (out, "}");
        }
      else
        fprintf (out, "\t∅");
    }

  fprintf (out, "\n");
}

static void
write_execution_report_extra_data (micro_benchmark_exec_report report,
                                   FILE *out)
{
  const char *full_name = micro_benchmark_exec_report_get_name (report);
  const size_t n_meters =
    micro_benchmark_exec_report_number_of_meters (report);
  if (n_meters > 0)
    {
      fprintf (out, SEPARATOR_DS_72);
      fprintf (out, _("%s: Extra data"), full_name);
    }
  for (size_t i = 0; i < n_meters; ++i)
    {
      micro_benchmark_report_extra_data extra =
        micro_benchmark_exec_report_get_extra_data (report, i);
      if (extra.name)
        fprintf (out, _("Start of %s\n"), extra.name);
      for (size_t j = 0; j < extra.size; ++j)
        fprintf (out, _("%s=\"%s\"\n"), extra.keys[j], extra.values[j]);
      if (extra.name)
        fprintf (out, _("End of %s\n"), extra.name);
    }
  if (n_meters > 0)
    fprintf (out, SEPARATOR_DS_72);
}

static void
write_test_report (micro_benchmark_test_report test, FILE *out,
                   const micro_benchmark_output_values *values)
{
  size_t n_cases = micro_benchmark_test_report_get_num_executions (test);
  if (n_cases > 1)
    fprintf (out, "%*s\n", test_name_size,
             micro_benchmark_test_report_get_name (test));

  for (size_t i = 0; i < n_cases; ++i)
    {
      micro_benchmark_exec_report e =
        micro_benchmark_test_report_get_exec_report (test, i);
      write_execution_report (e, out, values);
    }
  if (n_cases > 1)
    fprintf (out, SEPARATOR_DS_72);
}

static void
write_test_report_extra_data (micro_benchmark_test_report test, FILE *out)
{
  size_t n_cases = micro_benchmark_test_report_get_num_executions (test);
  if (n_cases > 1)
    fprintf (out, "%s\n", micro_benchmark_test_report_get_name (test));

  for (size_t i = 0; i < n_cases; ++i)
    {
      micro_benchmark_exec_report e =
        micro_benchmark_test_report_get_exec_report (test, i);
      write_execution_report_extra_data (e, out);
    }
  if (n_cases > 1)
    fprintf (out, SEPARATOR_DS_72);
}

void
micro_benchmark_report_write_console_ (micro_benchmark_report report,
                                       FILE *out,
                                       const micro_benchmark_output_values *v)
{
  MB_TRACE ("Writing console report %p.\n", report);
  assert (report);
  assert (out);
  assert (v);
  const size_t start = v->self_test ? 0 : 1;
  const char *suite_name = micro_benchmark_report_get_name (report);
  size_t n_tests = micro_benchmark_report_get_number_of_tests (report);
  write_header (out, v, suite_name, n_tests);

  for (size_t i = start; i < n_tests; ++i)
    {
      micro_benchmark_test_report r =
        micro_benchmark_report_get_test_report (report, i);
      write_test_report (r, out, v);
    }

  if (v->extra_data)
    for (size_t i = start; i < n_tests; ++i)
      {
        micro_benchmark_test_report r =
          micro_benchmark_report_get_test_report (report, i);
        write_test_report_extra_data (r, out);
      }
}
